/*
 ****************************************************************************
 * Copyright 2001-2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: mkivnet.c,v 1.6 2004/06/19 05:47:55 lreiter Exp $
 ***************************************************************************
 * This is part of the mki module that handles all of the linux specific
 * structures and function calls for virtual network (vnet).
 ***************************************************************************
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>
#include <linux/skbuff.h>
#include <linux/kernel.h>
#include <linux/smp_lock.h>
#include <linux/smp.h>
#include <asm/checksum.h>

#include "vneteth.h"
#include "vnetint-pub.h"
#include "mkifunc.h"

//#define DBGVNET(str1) printk str1
#define DBGVNET(str1)

spinlock_t vnet_inter_lock = SPIN_LOCK_UNLOCKED;
#define DECLARELOCKSAVE(flags) unsigned long flags=0
#define LOCKINTERCEPTOR(flags) spin_lock_irqsave(&vnet_inter_lock, flags)
#define UNLOCKINTERCEPTOR(flags) spin_unlock_irqrestore(&vnet_inter_lock, flags)

static struct sk_buff *mkia_checksum_help(struct sk_buff *skb)
{
	unsigned int csum;
	int offset = skb->h.raw - skb->data;

	if (offset > (int)skb->len)
		BUG();
	csum = skb_checksum(skb, offset, skb->len-offset, 0);

	offset = skb->tail - skb->h.raw;
	if (offset <= 0)
		BUG();
	if (skb->csum + 2 > offset)
		BUG();

	*(u16*)(skb->h.raw + skb->csum) = csum_fold(csum);
	skb->ip_summed = CHECKSUM_NONE;
	return skb;
}

int 
mkia_ntohs(int protocol)
{
	return ntohs(protocol);
}


int *mkia_vni_vnetdebug = 0;
void (*mkia_vni_logfunc)(char *, ...) = 0;

extern unsigned short eth_type_trans(struct sk_buff *skb, struct net_device *dev);
int mkia_vnetint_incoming(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt);
int mkia_vnetint_outgoing(struct sk_buff *skb, struct net_device *dev);

struct notifier_block mkia_nb;
int mkia_interceptor_count;
struct vnetint_ifdev            *mkia_vnetint_ifdev_base;
struct vnetint_proto            *mkia_vnetint_proto_base;
struct vnetint_pcb              *mkia_vnetint_inter_base;

/***************************/
/* VNET interceptor code   */
/***************************/

/*********************************/
/* dummy incoming packet handler */
int
mkia_vnetint_dummy(struct sk_buff *skb, 
			  struct net_device *dev, 
			  struct packet_type *pt)
{
	kfree_skb(skb);
	return 0;
}

struct vnetint_proto *
mkia_vnetint_find_proto(int type)
{
	struct vnetint_proto *vp;
	
	for (vp = mkia_vnetint_proto_base; vp != NULL; vp = vp->next)
		if (vp->ptype->type == type)
			break;
	
	return vp;
}


/* Add a protocol to the list. */
struct vnetint_proto *
mkia_vnetint_add_proto(int protocol)
{
	int type = ntohs(protocol);
	struct packet_type *pt, *node;
	struct vnetint_proto *vp;
	DECLARELOCKSAVE(flags);
		
	LOCKINTERCEPTOR(flags);
	if ( (vp = mkia_vnetint_find_proto(type)) ) {
		vp->count++;
		UNLOCKINTERCEPTOR(flags);
		return vp;
	}
	UNLOCKINTERCEPTOR(flags);
			
	if (!(vp = kmalloc(sizeof(struct vnetint_proto), GFP_KERNEL))) {
		return NULL;
	}
	vp->real_recv = NULL;
	
	if (!(pt = kmalloc(sizeof(struct packet_type), GFP_KERNEL))) {
		kfree(vp);
		return NULL;
	}
	
	pt->type = type;
	pt->func = mkia_vnetint_dummy;
	pt->dev = NULL;
	
	dev_add_pack(pt);

	LOCKINTERCEPTOR(flags);
	for (node = container_of(pt->list.next, struct packet_type, list);
	     ((node != NULL) && (node != pt));
	     node = container_of(node->list.next, struct packet_type, list)) {
		if (node->type == type && node->dev == NULL) {
			vp->ptype = node;
			vp->real_recv = node->func;
			node->func = mkia_vnetint_incoming;	
			DBGVNET(("vnetint_add_proto: found the real protocol "
				"handler\n"));
			break;
		}
	}
	if ((node == NULL) || (node == pt)) {
		pt->func = mkia_vnetint_incoming;
		vp->ptype = pt;
	}
	vp->protocol = protocol;
	vp->next = mkia_vnetint_proto_base;
	vp->count = 1;
	mkia_vnetint_proto_base = vp;
	UNLOCKINTERCEPTOR(flags);

	/* Remove the dummy we added, if it was extra */
	if ((node != NULL) && (node != pt)) {
		dev_remove_pack(pt);
		kfree(pt);
	}

	return vp;
}

/* Remove a protocol from the list. */
void
mkia_vnetint_rmv_proto(int protocol)
{
	struct vnetint_proto *vp, **pnode;
	struct packet_type *pt, *node;
	struct packet_type mypt;
	DECLARELOCKSAVE(flags);
	
	DBGVNET(("vnetint_rmv_proto: protocol type = 0x%0x\n", protocol));

	LOCKINTERCEPTOR(flags);
	for (pnode = &mkia_vnetint_proto_base; (*pnode) != NULL;
	     pnode = &((*pnode)->next)) {
		vp = (*pnode);
		if (vp->protocol != protocol)
			continue;

		vp->count--;
		if (vp->count > 0) {
			UNLOCKINTERCEPTOR(flags);
			return;
		}
		*pnode = (*pnode)->next;
		UNLOCKINTERCEPTOR(flags);
		if (vp->real_recv != NULL) {
			/* Check that the protocol is still registered */
			pt = &mypt;
			pt->type = htons(protocol);
			pt->func = mkia_vnetint_dummy;
			pt->dev = NULL;
			dev_add_pack(pt);
			for (node = container_of(pt->list.next,
						 struct packet_type, list);
			     ((node != NULL) && (node != pt));
			     node = container_of(node->list.next,
						 struct packet_type, list)) {

				if ((node == vp->ptype) &&
				    (node->func == mkia_vnetint_incoming)) {
					node->func = vp->real_recv;
					DBGVNET(("vnetint_rmv_proto: restored "
					    "the real protocol handler\n"));
				}
			}
			dev_remove_pack(pt);
		} else if (vp->ptype) {
			dev_remove_pack(vp->ptype);
			kfree(vp->ptype);
		}
		kfree(vp);
		return;
	}
	UNLOCKINTERCEPTOR(flags);
	
	DBGVNET(("vnetint_rmv_proto: protocol 0x%x not found.\n", protocol));
}

/* Find device from list - call with lock set */
struct vnetint_ifdev *
mkia_vnetint_find_dev(char *ifname)
{
	struct vnetint_ifdev *vd;
	
	for (vd = mkia_vnetint_ifdev_base; vd != NULL; vd = vd->next) {
		DBGVNET(("vnetint_find_dev: comparing %s with %s\n", vd->ifname, ifname));
		if (strcmp(vd->ifname, ifname) == 0) 
			break;
	}
	
	return vd;
}

/* FastFind device from list - call with lock set */
struct vnetint_ifdev *
mkia_vnetint_fastfind_dev(struct net_device *dev)
{
	struct vnetint_ifdev *vd;
	
	for (vd = mkia_vnetint_ifdev_base; vd != NULL; vd = vd->next)
		if (vd->dev == dev)
			break;
	
	return vd;
}

/* Add a device to the list. */
struct vnetint_ifdev *
mkia_vnetint_add_dev(char *ifname)
{
	struct vnetint_ifdev *vd;
	struct net_device *dev = dev_get_by_name(ifname);
	DECLARELOCKSAVE(flags);
	
	DBGVNET(("vnetint_add_dev: device = %s\n", ifname));

	if (!dev) {
		DBGVNET(("vnetint_add_dev: no such device %s\n", ifname));
		return NULL;
	}
	
	LOCKINTERCEPTOR(flags);
	if ( (vd = mkia_vnetint_find_dev(ifname)) ) {
		vd->count++;
		UNLOCKINTERCEPTOR(flags);
		dev_put(dev);
		return vd;
	}
	UNLOCKINTERCEPTOR(flags);
			
	if (!(vd = kmalloc(sizeof(struct vnetint_ifdev), GFP_KERNEL))) {
		dev_put(dev);
		return NULL;
	}
	
	LOCKINTERCEPTOR(flags);
	vd->dev = dev;
	strncpy(vd->ifname, ifname, MAXIFNAMELEN);
	vd->real_xmit = dev->hard_start_xmit;
	dev->hard_start_xmit = mkia_vnetint_outgoing;
	vd->next = mkia_vnetint_ifdev_base;
	vd->count = 1;
	mkia_vnetint_ifdev_base = vd;
	UNLOCKINTERCEPTOR(flags);
	
	return vd;
}


/* Remove a device from the list. */
void
mkia_vnetint_rmv_dev(char *ifname)
{
	
	struct vnetint_ifdev *vd, **pnode;
	struct net_device *ndp;
	DECLARELOCKSAVE(flags);
		
	DBGVNET(("vnetint_rmv_dev: device = %s\n", ifname));
	
	LOCKINTERCEPTOR(flags);
	for (pnode = &mkia_vnetint_ifdev_base; (*pnode) != NULL;
	     pnode = &((*pnode)->next)) {
		vd = (*pnode);
		if (strcmp(vd->ifname, ifname) == 0) {
			vd->count--;
			if (vd->count > 0) {
				UNLOCKINTERCEPTOR(flags);
				return;
			}
			ndp = (struct net_device *) (vd->dev);
			if ((ndp != NULL) &&
			    (ndp->hard_start_xmit == mkia_vnetint_outgoing)) {
				ndp->hard_start_xmit = vd->real_xmit;
			}
			*pnode = (*pnode)->next;
			UNLOCKINTERCEPTOR(flags);
			if (vd->dev != NULL) {
				dev_put(vd->dev);
			}
			kfree(vd);
			return;
		}
	}	
	UNLOCKINTERCEPTOR(flags);
	
	DBGVNET(("vnetint_rmv_dev: dev %s not found.\n", ifname));
}



/* Called when network device changes state */
int
mkia_vnetint_dev_notifier(struct notifier_block *self, 
				 unsigned long status, 
				 void *ptr)
{
	
	struct net_device *dev = (struct net_device *)ptr;

	struct vnetint_ifdev *vd;
	DECLARELOCKSAVE(flags);

	DBGVNET(("vnetint_dev_notifier: Device %s changed state\n", dev->name));
		
	LOCKINTERCEPTOR(flags);
	switch (status) {
		case NETDEV_CHANGE:
			DBGVNET(("vnetint_dev_notifier: NETDEV_CHANGE\n"));
			break;
		case NETDEV_UP:
			DBGVNET(("vnetint_dev_notifier: NETDEV_UP\n"));
			vd = mkia_vnetint_fastfind_dev(dev);
			if (vd) {
				vd->real_xmit = dev->hard_start_xmit;
				dev->hard_start_xmit = mkia_vnetint_outgoing;
			}
			else {
				DBGVNET(("vnetint_dev_notifier: "
				    "dev %s not found\n", dev->name));
			}
			break;		
		case NETDEV_DOWN:
			DBGVNET(("vnetint_dev_notifier: NETDEV_DOWN\n"));
			vd = mkia_vnetint_fastfind_dev(dev);
			if (vd) {
				DBGVNET(("vnetint_dev_notifier: "
				    "found device\n"));
				if (dev->hard_start_xmit ==
				    mkia_vnetint_outgoing) {
					DBGVNET(("vnetint_dev_notifier: "
					    "restore real_xmit\n"));
					dev->hard_start_xmit = vd->real_xmit;
				}
				vd->real_xmit = NULL;
			}
			else {
				DBGVNET(("vnetint_dev_notifier: "
				    "dev %s not found\n", dev->name));
			}
			break;
		case NETDEV_CHANGEMTU:
			DBGVNET(("vnetint_dev_notifier: NETDEV_CHANGEMTU\n"));
			break;
		case NETDEV_CHANGEADDR:
			DBGVNET(("vnetint_dev_notifier: NETDEV_CHANGEADDR\n"));
			break;
		case NETDEV_CHANGENAME:
			DBGVNET(("vnetint_dev_notifier: NETDEV_CHANGENAME\n"));
			break;
		case NETDEV_REGISTER:
			DBGVNET(("vnetint_dev_notifier: NETDEV_REGISTER\n"));
			vd = mkia_vnetint_find_dev(dev->name);
			if (vd) {
				vd->dev = dev_get_by_name(dev->name);
			}
			break;
		case NETDEV_UNREGISTER:
			DBGVNET(("vnetint_dev_notifier: NETDEV_UNREGISTER\n"));
			vd = mkia_vnetint_find_dev(dev->name);
			if (vd) {
				vd->dev = NULL;
				if (dev->hard_start_xmit ==
				    mkia_vnetint_outgoing) {
					dev->hard_start_xmit = vd->real_xmit;
				}
				vd->real_xmit = NULL;
				dev_put(dev);
			}
			else {
				DBGVNET(("vnetint_dev_notifier: "
				    "dev %s not found\n", dev->name));
			}
			break;
		default:
			DBGVNET(("vnetint_dev_notifier: "
			    "Unknown device status\n"));
			break;
	}
	UNLOCKINTERCEPTOR(flags);
	
	DBGVNET(("vnetint_dev_notifier: done\n"));

	return 0;
}


/* Find interceptor on list. - call with lock set */
struct vnetint_pcb *
mkia_vnetint_find_inter(int protocol, char *ifname)
{
	struct vnetint_pcb *vi;
	
	for (vi = mkia_vnetint_inter_base; vi != NULL; vi = vi->next) {
		if ((strncmp(vi->ifdev->ifname, ifname, MAXIFNAMELEN) == 0) &&
			(vi->proto->protocol == protocol)) {
			break;
		}
	}
	
	return vi;
}


/* FastFind interceptor on list. - call with lock set */
static inline struct vnetint_pcb *
mkia_vnetint_fastfind_inter(int type, struct net_device *dev)
{
	struct vnetint_pcb *vi;
	
	for (vi = mkia_vnetint_inter_base; vi != NULL; vi = vi->next) {
		if ((vi->ifdev->dev == dev) &&
		    (vi->proto->ptype->type == type)) {
			break;
		}
	}
	
	return vi;
}


/* Add a an interceptor to the list */
int
mkia_vnetint_add_interceptor(struct vnetint_filter *filter)
{
	
	struct vnetint_pcb *vi;
	DECLARELOCKSAVE(flags);
	
	DBGVNET(("vnetint_add_interceptor: device = %8s "
	    "protocol type = 0x%0x filter_func = 0x%p\n", filter->ifname,
	    filter->protocol, filter->intercept_func));

	if (filter->intercept_func == NULL) {
		return -EINVAL;
	}
	
	if (mkia_interceptor_count == MAX_INTERCEPTORS) {
		return -ENOENT;
	}

	if ((vi = mkia_vnetint_find_inter(filter->protocol, filter->ifname))) {
		/*
		 * An interceptor aready exists for this dev/proto pair.
		 * Let the caller have the interceptor reference but return
		 * EEXIST to warn that this interceptor was aready defined.
		 */
		LOCKINTERCEPTOR(flags);
		vi->count++; /* Increment the interceptor usage count */
		filter->handle = vi;
		UNLOCKINTERCEPTOR(flags);
		return -EEXIST;
	}
				
	if (!(vi = kmalloc(sizeof(struct vnetint_pcb), GFP_KERNEL))) {
		return -ENOMEM;
	}
	
	vi->ifdev = mkia_vnetint_add_dev(filter->ifname);
	if (vi->ifdev == NULL) {
		kfree(vi);
		return -ENXIO;
	}
	
	vi->proto = mkia_vnetint_add_proto(filter->protocol);
	if (vi->proto == NULL) {
		mkia_vnetint_rmv_dev(filter->ifname);
		kfree(vi);
		return -ENXIO;
	}
	
	vi->filter_func = filter->intercept_func;
	memset(&(vi->status), 0, sizeof(struct vnetint_status));
	vi->status.flags = 0;
	memcpy(&(vi->status.PhysMacAddr),
	    ((struct net_device *) vi->ifdev->dev)->dev_addr, ETH_ALEN);
	vi->status.flags |= VNETINTF_MACSET;
	LOCKINTERCEPTOR(flags);
	vi->next = mkia_vnetint_inter_base;
	vi->count = 1;
	mkia_vnetint_inter_base = vi;
	mkia_interceptor_count++;
	filter->handle = vi;
	UNLOCKINTERCEPTOR(flags);
	
	return(0);
}


/* Remove an interceptor from the list */
void
mkia_vnetint_rmv_interceptor(struct vnetint_filter *filter)
{
	
	char *ifname;
	int protocol;
	struct vnetint_pcb *vi, **pnode;
	DECLARELOCKSAVE(flags);
	
	if (filter == NULL) {
		DBGVNET(("vnetint_rmv_interceptor: remove all interceptors\n"));
		LOCKINTERCEPTOR(flags);
		for (pnode = &mkia_vnetint_inter_base; (*pnode) != NULL;
		     pnode = &((*pnode)->next)) {
			ifname = (*pnode)->ifdev->ifname;
			protocol = (*pnode)->proto->protocol;
			DBGVNET(("vnetint_rmv_interceptor: device = %s, "
			    "protocol type = 0x%0x\n", ifname, protocol));
			mkia_interceptor_count--;
			UNLOCKINTERCEPTOR(flags);
			mkia_vnetint_rmv_dev(ifname);
			mkia_vnetint_rmv_proto(protocol);
			kfree((*pnode));
			LOCKINTERCEPTOR(flags);
		}
		UNLOCKINTERCEPTOR(flags);
		return;
	}
	
	vi = filter->handle;
	if (vi) {
		DBGVNET(("vnetint_rmv_interceptor: device = %s, "
		    "protocol type = 0x%0x\n", filter->ifname,
		    filter->protocol));
		LOCKINTERCEPTOR(flags);
		for (pnode = &mkia_vnetint_inter_base; (*pnode) != NULL;
		     pnode = &((*pnode)->next)) {
			if ((*pnode) == vi) {
				vi->count--;
				if ( vi->count > 0 ) {
					UNLOCKINTERCEPTOR(flags);
					return;
				}
				*pnode = (*pnode)->next;
				mkia_interceptor_count--;
				UNLOCKINTERCEPTOR(flags);
				mkia_vnetint_rmv_dev(filter->ifname);
				mkia_vnetint_rmv_proto(filter->protocol);
				kfree(vi);
				return;
			}
		}
		UNLOCKINTERCEPTOR(flags);

		
	}
	
	DBGVNET(("vnetint_rmv_interceptor: no interceptor found\n"));
	
	return;
}

/* Incoming Packet Handler */
int
mkia_vnetint_incoming(struct sk_buff *skb, struct net_device *dev,
	struct packet_type *pt)
{
	
	struct vnetint_proto *vp;
	struct vnetint_pcb *vi;
	filter_func_t filter_func = 0;
	recv_func_t real_recv = 0;
	int consumed;
	DECLARELOCKSAVE(flags);

	DBGVNET(("vnetint_incoming: got packet from dev %s for "
	    "protocol 0x%04x\n", dev->name, ntohs(pt->type)));
	
	LOCKINTERCEPTOR(flags);
	vp = mkia_vnetint_find_proto(pt->type);
	if (vp == NULL) {
		/*
		 * This should never happen but just to play safe ...
		 * Drop the packet
		 */
		printk("vnetint_incoming: protocol 0x%04x not registered!!!\n",
			ntohs(pt->type));
		goto incoming_free_pkt;
	}
	real_recv = vp->real_recv;
	if ((skb->pkt_type & (PACKET_HOST | PACKET_OTHERHOST))
	    		== (PACKET_HOST | PACKET_OTHERHOST)) {
		DBGVNET(("vnetint_incoming: ignore pkt from filter)\n"
			"- send it on through\n"));
		skb->pkt_type &= ~PACKET_OTHERHOST;
		goto pass_to_original_handler;
	}

	vi = mkia_vnetint_fastfind_inter(pt->type, dev);
	if (vi == NULL) {
		DBGVNET(("vnetint_incoming: no interceptor for "
		    "dev %s, protocol 0x%x\n", dev->name, ntohs(pt->type)));
		goto pass_to_original_handler;
	}

	if (! VNETINT_ACT(vi)) {
		DBGVNET(("vnetint_incoming: interceptor not active\n"));
		goto pass_to_original_handler;
	}

	vi->status.NumPktsUp++;
	filter_func = vi->filter_func;
	if (filter_func == NULL) {
		DBGVNET(("vnetint_incoming: missing filter_func\n"));
		goto pass_to_original_handler;
	}

	UNLOCKINTERCEPTOR(flags);

	consumed = filter_func(skb->mac.raw,
				skb->len + dev->hard_header_len, skb);

	LOCKINTERCEPTOR(flags);

	if (consumed) {
		vi->status.NumPktsConsumed++;
		goto incoming_free_pkt;
	}

	/*
	 * Fall through and send the packet to the original handler since
	 * the filter did not full consume it.
	 */
pass_to_original_handler:
	/* Pass packet to original protocol handler */
	if (!real_recv) {
		DBGVNET(("vnetint_incoming: no protocol - "
		    "dropped packet\n"));
		goto incoming_free_pkt;
	}
	UNLOCKINTERCEPTOR(flags);
	return (*real_recv)(skb, dev, pt);

incoming_free_pkt:
	UNLOCKINTERCEPTOR(flags);
	kfree_skb(skb);
	return 0;
}


/* Outgoing Packet Handler */
int
mkia_vnetint_outgoing(struct sk_buff *skb, struct net_device *dev)
{
	
	struct vnetint_ifdev *vd;
	struct vnetint_pcb *vi;
	xmit_func_t real_xmit = 0;
	ETH_HDR_T *eth;
	unsigned short protocol;
	int consumed;
	DECLARELOCKSAVE(flags);

	/*
	 * We can't rely on skb->protocol to hold a valid protocol
	 * type so go look in the ethernet header.
	 */
	eth = (ETH_HDR_T *)skb->data;
	protocol = eth->type;

	DBGVNET(("vnetint_outgoing: got packet for dev %s from protocol "
		 "0x%04x\n", dev->name, ntohs(protocol)));

	LOCKINTERCEPTOR(flags);
	
	vd = mkia_vnetint_fastfind_dev(dev);
	if (vd == NULL) {
		/* This should never happen but just to play safe ... */
		/* Drop the packet */
		printk("vnetint_outgoing: device %p not registered!!!\n", dev);
		goto outgoing_free_pkt;
	}
	real_xmit = vd->real_xmit;
	
	if ((skb->pkt_type & (PACKET_OUTGOING | PACKET_OTHERHOST))
	    		== (PACKET_OUTGOING | PACKET_OTHERHOST)) {
		DBGVNET(("vnetint_outgoing: ignore pkt from filter "
			"- send it on through\n"));
		skb->pkt_type &= ~PACKET_OTHERHOST;
		goto pass_to_device;
	}
	vi = mkia_vnetint_fastfind_inter(protocol, dev);
	if (vi == NULL) {
		DBGVNET(("vnetint_outgoing: no interceptor for dev %s, "
			"protocol 0x%04x\n", dev->name, ntohs(protocol)));
		goto pass_to_device;
	}
	if (! VNETINT_ACT(vi)) {
		DBGVNET(("vnetint_outgoing: interceptor not active\n"));
		goto pass_to_device;
	}

	vi->status.NumPktsDown++;
	if (vi->filter_func) {
		/* mac.raw is not set by protocol driver */
		skb->mac.raw = skb->data;

		UNLOCKINTERCEPTOR(flags);
		consumed = vi->filter_func(skb->data, skb->len, skb);
		LOCKINTERCEPTOR(flags);
		if (consumed) {
			DBGVNET(("vnetinf_outgoing: filter consumed "
				"packet\n"));
			vi->status.NumPktsConsumed++;
			goto outgoing_free_pkt;
		}
	}
	
pass_to_device:
	DBGVNET(("vnetint_outgoing: give packet to real NIC xmitfunc\n"));
	/* Pass packet to the network device */
	if (!real_xmit) {
		printk("vnetint_outgoing: bad device - dropped packet\n");	
		goto outgoing_free_pkt;
		return 0;
	}
	UNLOCKINTERCEPTOR(flags);
	return (*real_xmit)(skb, dev);

outgoing_free_pkt:
	UNLOCKINTERCEPTOR(flags);
	kfree_skb(skb);
	return 0;
}


/* Convert Ethernet address to printable (loggable) representation. */
char *
mkia_ether_sprintf(ETH_ADDR_T *addr)
{
	register int		i;
	static char	 etherbuf[18];
	register char  *cp = etherbuf;
	register unsigned char  *ap = (unsigned char *)addr;
	static char	 digits[] = "0123456789abcdef";

	for (i = 0; i < 6; i++) {
		if (*ap > 16)
			*cp++ = digits[*ap >> 4];
		*cp++ = digits[*ap++ & 0xf];
		*cp++ = ':';
	}
	*--cp = 0;
	return (etherbuf);
}

/* Interceptor process packet */
void
vnetint_ProcessPacket(char *pPacket, 
		      unsigned size, 
		      int direction, 
		      void *handle)
{
	
	struct sk_buff *skb = dev_alloc_skb(size + 2);
	struct vnetint_pcb *vi = (struct vnetint_pcb *)handle;
	struct net_device *dev = vi->ifdev->dev;
	DECLARELOCKSAVE(flags);
	
	DBGVNET(("vnetint_ProcessPacket: direction 0x%x:\n",direction));

	if (dev == NULL || skb == NULL) {
		/* No can do - drop packet */
		return;
	}

	skb->dev = dev;
	skb_reserve(skb, 2);	/* Align IP on 16 byte boundaries */
	memcpy(skb_put(skb, size), pPacket, size);

	/*
	 * It seems that net/core/dev.c excpects this field to be
	 * initialized, so we will force it in the same way the kernel
	 * does if it's not initialized.
	 *
	 * This prevents the "protocol buggy" messages that were showing
	 * up in the syslog.
	 */
	if (skb->nh.raw < skb->data || skb->nh.raw > skb->tail) {
		skb->nh.raw = skb->data;
	}

	LOCKINTERCEPTOR(flags);
	vi->status.NumPktsFromVnet++;
	if (VNETINT_ACT(vi)) {
		UNLOCKINTERCEPTOR(flags);
		/*
		 * NOTE: in the following code, use of PACKET_OTHERHOST
		 * is nonsensical. This fact is being used to mark the
		 * packet as having originated in the VNET so that the
		 * packet can be recognized and ignored by the interceptor
		 * incoming and outgoing packet handlers.
		 */
		if (direction == DOWNSTREAM) {
			/* Pass packet to the network device outbound queue */
			DBGVNET(("vnetint_ProcessPacket: sending packet down "
			    "to net device\n"));
			skb->pkt_type = (PACKET_OUTGOING | PACKET_OTHERHOST);

			dev_queue_xmit(skb);
		}
		else {
			/* Pass packet to the network device inbound queue */
			DBGVNET(("vnetint_ProcessPacket: sending packet up "
			    "to protocol\n"));
			skb->protocol = eth_type_trans(skb, dev);
			skb->pkt_type = (PACKET_HOST | PACKET_OTHERHOST);
			netif_rx(skb);
		}
		DBGVNET(("vnetint_ProcessPacket done\n"));
		return;
	}
	else {
		DBGVNET(("vnetint_ProcessPacket: Dropped packet - "
		    "interceptor down\n"));
	}
	UNLOCKINTERCEPTOR(flags);

	kfree_skb(skb);
}
EXPORT_SYMBOL(vnetint_ProcessPacket);

/* Interceptor control status */
void
vnetint_CtrlStatus(unsigned MsgType, struct vnetint_filter *filter)
{
	
	struct vnetint_pcb *vi = filter->handle;
	struct net_device *ndp;
	DECLARELOCKSAVE(flags);
	
	DBGVNET(("vnetint_CtrlStatus: MsgType=%x, filter=0x%p\n",
	    MsgType, filter));
	LOCKINTERCEPTOR(flags);
	switch (MsgType) {
		case INTERCEPTOR_GOACTIVE:
			DBGVNET(("vnetint_CtrlStatus: GOACTIVE\n"));
			vi->status.flags |= VNETINTF_ACTIVE;
			break;
		case INTERCEPTOR_GOPASSIVE:
			DBGVNET(("vnetint_CtrlStatus: GOPASSIVE\n"));
			vi->status.flags &= ~VNETINTF_ACTIVE;
			break;
		case INTERCEPTOR_GETSTATS:
			DBGVNET(("vnetint_CtrlStatus: GETSTATS\n"));
			if (vi->ifdev->dev) {
				ndp = ((struct net_device *) vi->ifdev->dev);
				memcpy(&(vi->status.PhysMacAddr),
				    ndp->dev_addr, ETH_ALEN);
			}
			else {
				memset(&(vi->status.PhysMacAddr), 0, ETH_ALEN);
			}
			memcpy(&(filter->status), &(vi->status),
			    sizeof(struct vnetint_status));
			break;
		case INTERCEPTOR_SETFILTER:
			DBGVNET(("vnetint_CtrlStatus: SETFILTER "));
			UNLOCKINTERCEPTOR(flags);
			if (filter->intercept_func) {
				DBGVNET(("- add interceptor\n"));
				mkia_vnetint_add_interceptor(filter);
			}
			else {
				DBGVNET(("- remove interceptor\n"));
				mkia_vnetint_rmv_interceptor(filter);
			}
			return;
		default:
			break;
	}
	UNLOCKINTERCEPTOR(flags);
	return;
}
EXPORT_SYMBOL(vnetint_CtrlStatus);

void
vnetint_setdebug(int *debug_mask_ptr, void (*debug_log_func)(char *, ...))
{
	mkia_vni_vnetdebug = debug_mask_ptr;
	mkia_vni_logfunc = debug_log_func;
}
EXPORT_SYMBOL(vnetint_setdebug);

/* Interceptor copy packet*/
unsigned
vnetint_CopyPacket(void *cookie, 
		   unsigned offset, 
		   char *dest, 
		   unsigned maxpktsz)
{
	struct sk_buff *skb = cookie;
	unsigned data_len = skb->tail - skb->mac.raw;
	
	if (skb->ip_summed == CHECKSUM_HW) {
		skb = mkia_checksum_help(skb);
		if (skb == NULL) {
			printk("mki-adapter: skb_checksum_help failed\n");
			return 0;
		}
	}

	DBGVNET(("vnetint_CopyPacket: cookie = %p, offset = %d, dest = %p, "
	    "maxpktsz = %d\n", cookie, offset, dest, maxpktsz));
	DBGVNET(("vnetint_CopyPacket: nr_frags %d, frag_list %p\n",
	     skb_shinfo(skb)->nr_frags, skb_shinfo(skb)->frag_list));

	data_len -= offset;
	if (data_len > maxpktsz)
		data_len = maxpktsz;
	memcpy(dest, skb->mac.raw + offset, data_len);
	return data_len;
}
EXPORT_SYMBOL(vnetint_CopyPacket);

int
mkia_init_vnetint(void)
{
	mkia_vnetint_ifdev_base = NULL;
	mkia_vnetint_proto_base = NULL;
	mkia_vnetint_inter_base = NULL;
	mkia_interceptor_count = 0;
	
	mkia_nb.notifier_call = mkia_vnetint_dev_notifier;
	mkia_nb.next = NULL;
	mkia_nb.priority = 0;

	register_netdevice_notifier(&mkia_nb);
	
	return 0;
}
EXPORT_SYMBOL(mkia_init_vnetint);

void
mkia_cleanup_vnetint(void)
{
	unregister_netdevice_notifier(&mkia_nb);
	mkia_vnetint_rmv_interceptor(NULL);
}
EXPORT_SYMBOL(mkia_cleanup_vnetint);

